__all__ = [
    'convert_geojson',
    'legacy_pairs_to_geojson',
    'geojson_to_legacy_pairs'
]

import functools
from datetime import datetime, date, time
from typing import List

import bson

from mongoengine_migrate.exceptions import MigrationError, InconsistencyError
from mongoengine_migrate.mongo import (
    check_empty_result,
    mongo_version
)
from mongoengine_migrate.updater import ByPathContext, ByDocContext, DocumentUpdater

#: GeoJSON field convertions in order of increasing the nested array
#  depth in `coordinates` subfield.
#: The simplest is Point (`coordinates` contains array of numbers),
#: the most complex is MultiPolygon (`coordinates` consists of array
#: of arrays of arrays of arrays of numbers).
__CONVERTIONS = (
    ("Point", ),
    ("MultiPoint", "LineString"),
    ("Polygon", "MultiLineString"),
    ("MultiPolygon", )
)


def convert_geojson(updater: DocumentUpdater, from_type: str, to_type: str):
    """Convert GeoJSON object from one type to another"""
    from_ind, to_ind = None, None
    if updater.migration_policy.name == 'strict':
        __check_geojson_objects(updater, [from_type, to_type])
        # There could be legacy coordinate pair arrays or GeoJSON objects
        __check_legacy_point_coordinates(updater)
        __check_value_types(updater, ['object', 'array'])

    if from_type == to_type:
        return

    for ind, convertion in __CONVERTIONS:
        if from_type in convertion:
            from_ind = ind
        if to_type in convertion:
            to_ind = ind

    if from_ind is None or to_ind is None:
        raise MigrationError(f"Unknown geo field type. Was requested: {from_type}, {to_type}")

    depth = abs(from_ind - to_ind)
    if from_ind <= to_ind:
        __increase_geojson_nesting(updater, from_type, to_type, depth)
    else:
        __decrease_geojson_nesting(updater, from_type, to_type, depth)


def legacy_pairs_to_geojson(updater: DocumentUpdater, to_type: str):
    """Convert legacy coordinate pairs to GeoJSON objects of given type"""
    def by_doc(ctx: ByDocContext):
        doc = ctx.document
        if isinstance(doc.get(updater.field_name), (list, tuple)):
            doc[updater.field_name] = {'type': 'Point', 'coordinates': doc[updater.field_name]}

    if updater.migration_policy.name == 'strict':
        __check_geojson_objects(updater, ['Point', to_type])
        __check_legacy_point_coordinates(updater)
        __check_value_types(updater, ['object', 'array'])

    updater.update_by_document(by_doc)
    convert_geojson(updater, 'Point', to_type)


def geojson_to_legacy_pairs(updater: DocumentUpdater, from_type: str):
    """Convert GeoJSON objects of given type to legacy coordinate pairs"""
    def by_doc(ctx: ByDocContext):
        doc = ctx.document
        if isinstance(doc.get(updater.field_name), dict):
            if 'Point' in doc[updater.field_name]:
                doc[updater.field_name] = doc[updater.field_name].get('coordinates')

    if updater.migration_policy.name == 'strict':
        __check_geojson_objects(updater, ["Point", from_type])
        __check_legacy_point_coordinates(updater)
        __check_value_types(updater, ['object', 'array'])

    convert_geojson(updater, from_type, 'Point')

    updater.update_by_document(by_doc)


def __increase_geojson_nesting(updater: DocumentUpdater,
                               from_type: str,
                               to_type: str,
                               depth: int = 1):
    """
    Wraps `coordinates` field into nested array on GeoJSON fields
    with given type.
    :param updater: DocumentUpdater object
    :param from_type: GeoJSON type to change
    :param to_type: this GeoJSON type will be set in changed records
    :param depth: nested array depth to wrap in
    :return:
    """
    assert depth > 0

    def by_doc(ctx: ByDocContext):
        doc = ctx.document
        if isinstance(doc.get(updater.field_name), dict):
            match = doc[updater.field_name].get('type') == from_type \
                    and doc[updater.field_name].get('coordinates')
            if match:
                doc[updater.field_name]['coordinates'] = functools.reduce(
                    lambda x, y: [x],
                    range(depth),
                    doc[updater.field_name].get('coordinates', [.0, .0])
                )

    updater.update_by_document(by_doc)


def __decrease_geojson_nesting(updater: DocumentUpdater,
                               from_type: str,
                               to_type: str,
                               depth: int = 1):
    """
    Extract the first element from nested arrays in `coordinates` field
    on GeoJSON fields with given type
    :param updater: DocumentUpdater object
    :param from_type: GeoJSON type to change
    :param to_type: this GeoJSON type will be set in changed records
    :param depth: nested array depth to extract from
    :return:
    """
    assert depth > 0

    def by_doc(ctx: ByDocContext):
        doc = ctx.document
        if isinstance(doc.get(updater.field_name), dict):
            match = doc[updater.field_name].get('type') == from_type \
                    and doc[updater.field_name].get('coordinates')
            if match:
                doc[updater.field_name]['coordinates'] = functools.reduce(
                    lambda x, y: x[0] if x and isinstance(x, (list, tuple)) else None,
                    range(depth),
                    doc[updater.field_name].get('coordinates', [.0, .0])
                )

    updater.update_by_document(by_doc)


@mongo_version(min_version='3.6')
def __check_geojson_objects(updater: DocumentUpdater, geojson_types: List[str]):
    """
    Check if all object values in field are GeoJSON objects of given
    types. Raise InconsistencyError if other objects found
    :param updater:
    :param geojson_types:
    :return:
    """
    def by_path(ctx: ByPathContext):
        fltr = {"$and": [
            {ctx.filter_dotpath: {"$ne": None}},
            *[{k: v} for k, v in ctx.extra_filter.items()],
            {f'{ctx.filter_dotpath}.type': {'$nin': geojson_types}},
            # $expr >= 3.6
            {"$expr": {"$eq": [{"$type": f'${ctx.filter_dotpath}'}, 'object']}}
        ]}
        check_empty_result(ctx.collection, ctx.filter_dotpath, fltr)

    def by_doc(ctx: ByDocContext):
        doc = ctx.document
        if updater.field_name in doc:
            f = doc[updater.field_name]
            valid = f is None or (isinstance(f, dict) and f.get('type') in geojson_types)
            if not valid:
                raise InconsistencyError(f"Field {updater.field_name} has wrong value {f!r} "
                                         f"(should be GeoJSON) in record {doc}")

    updater.update_combined(by_path, by_doc, False, False)


@mongo_version(min_version='3.6')
def __check_legacy_point_coordinates(updater: DocumentUpdater):
    """
    Check if all array values in field has legacy geo point
    coordinates type. Raise InconsistencyError if other arrays was found
    :param updater:
    :return:
    """
    def by_path(ctx: ByPathContext):
        fltr = {"$and": [
            {ctx.filter_dotpath: {"$ne": None}},
            *[{k: v} for k, v in ctx.extra_filter.items()],
            # $expr >= 3.6, $isArray >= 3.2
            {"$expr": {"$eq": [{"$isArray": f"${ctx.filter_dotpath}"}, True]}},
            {"$expr": {"$ne": [{"$size": f"${ctx.filter_dotpath}"}, 2]}},  # $expr >= 3.6
            # TODO: add element type check
        ]}
        check_empty_result(ctx.collection, ctx.filter_dotpath, fltr)

    def by_doc(ctx: ByDocContext):
        doc = ctx.document
        if updater.field_name in doc:
            f = doc[updater.field_name]
            valid = f is None or (isinstance(f, (list, tuple)) and len(f) == 2)
            if not valid:
                raise InconsistencyError(f"Field {updater.field_name} has wrong value {f!r} "
                                         f"(should be legacy geo point) in record {doc}")

    updater.update_combined(by_path, by_doc, False, False)


@mongo_version(min_version='3.6')
def __check_value_types(updater: DocumentUpdater, allowed_types: List[str]):
    """
    Check if given field contains only given types of value.
    Raise InconsistencyError if other value types was found
    :param updater:
    :param allowed_types:
    :return:
    """
    def by_path(ctx: ByPathContext):
        # Check for data types other than objects or arrays
        fltr = {"$and": [
            {ctx.filter_dotpath: {"$ne": None}},
            *[{k: v} for k, v in ctx.extra_filter.items()],
            # $expr >= 3.6, $type >= 3.4
            {"$expr": {"$not": [{"$in": [{"$type": f'${ctx.filter_dotpath}'}, allowed_types]}]}}
        ]}
        check_empty_result(ctx.collection, ctx.filter_dotpath, fltr)

    def by_doc(ctx: ByDocContext):
        # https://docs.mongodb.com/manual/reference/operator/aggregation/convert/
        type_map = {
            'double': float,
            'string': str,
            'objectId': bson.ObjectId,
            'bool': bool,
            'date': datetime,
            'int': int,
            'long': int,
            'decimal': float
        }
        assert set(allowed_types) < type_map.keys()

        doc = ctx.document
        if updater.field_name in doc:
            f = doc[updater.field_name]
            valid_types = tuple(type_map[t] for t in allowed_types)
            valid = f is None or isinstance(f, valid_types)
            if not valid:
                raise InconsistencyError(f"Field {updater.field_name} has wrong type of value "
                                         f"{f!r} (should be any of {valid_types}) in record {doc}")

    updater.update_combined(by_path, by_doc, False, False)



__all__ = [
    'build_actions_chain',
    'build_document_action_chain',
    'build_field_action_chain',
    'get_all_document_types'
]

import logging
from copy import copy
from typing import Iterable, Type

from dictdiffer import patch, diff

import mongoengine_migrate.flags as flags
from mongoengine_migrate.exceptions import ActionError
from mongoengine_migrate.schema import Schema
from .base import (
    actions_registry,
    BaseFieldAction,
    BaseDocumentAction,
    BaseAction
)

log = logging.getLogger('mongoengine-migrate')


def build_actions_chain(left_schema: Schema, right_schema: Schema) -> Iterable[BaseAction]:
    """
    Build full Action objects chain which suitable for such schema
    change.
    :param left_schema: current schema
    :param right_schema: schema collected from mongoengine models
    :return: iterable of Action objects
    """
    action_chain = []

    # Actions registry sorted by priority
    registry = list(sorted(actions_registry.values(), key=lambda x: x.priority))

    left_schema = copy(left_schema)
    document_types = get_all_document_types(left_schema, right_schema)
    for action_cls in registry:
        if issubclass(action_cls, BaseDocumentAction):
            new_actions = list(
                build_document_action_chain(action_cls, left_schema, right_schema, document_types)
            )
        elif issubclass(action_cls, BaseFieldAction):
            new_actions = list(
                build_field_action_chain(action_cls, left_schema, right_schema, document_types)
            )
        else:
            continue

        for action in new_actions:
            log.debug('> %s', action)
            try:
                left_schema = patch(action.to_schema_patch(left_schema), left_schema)
            except (TypeError, ValueError, KeyError) as e:
                raise ActionError(
                    f"Unable to apply schema patch of {action!r}. More likely that the "
                    f"schema is corrupted. You can use schema repair tools to fix this issue"
                ) from e
        action_chain.extend(new_actions)
        document_types = get_all_document_types(left_schema, right_schema)

    if right_schema != left_schema:
        log.error(
            'Schema is still not reached the target state after applying all actions. '
            'Changes left to make (diff): %s',
            list(diff(left_schema, right_schema))
        )
        raise ActionError('Could not reach target schema state after applying whole Action chain. '
                          'This could be a problem in some Action which does not process schema '
                          'properly or produces wrong schema diff. This is a programming error')

    return action_chain


def build_document_action_chain(action_cls: Type[BaseDocumentAction],
                                left_schema: Schema,
                                right_schema: Schema,
                                document_types: Iterable[str]) -> Iterable[BaseAction]:
    """
    Walk through schema changes, and produce chain of Action objects
    of given type which could handle schema changes from left to right
    :param action_cls: Action type to consider
    :param left_schema:
    :param right_schema:
    :param document_types: list of document types to inspect
    :return: iterable of suitable Action objects
    """
    for document_type in document_types:
        action_obj = action_cls.build_object(document_type, left_schema, right_schema)
        if action_obj is not None:
            try:
                left_schema = patch(action_obj.to_schema_patch(left_schema), left_schema)
            except (TypeError, ValueError, KeyError) as e:
                raise ActionError(
                    f"Unable to apply schema patch of {action_obj!r}. More likely that the "
                    f"schema is corrupted. You can use schema repair tools to fix this issue"
                ) from e

            yield action_obj


def build_field_action_chain(action_cls: Type[BaseFieldAction],
                             left_schema: Schema,
                             right_schema: Schema,
                             document_types: Iterable[str]) -> Iterable[BaseAction]:
    """
    Walk through schema changes, and produce chain of Action objects
    of given type which could handle schema changes from left to right
    :param action_cls: Action type to consider
    :param left_schema:
    :param right_schema:
    :param document_types: list of document types to inspect
    :return: iterable of suitable Action objects
    """
    for document_type in document_types:
        # Take all fields to detect if they created, changed or dropped
        all_fields = left_schema.get(document_type, {}).keys() | \
                     right_schema.get(document_type, {}).keys()
        for field in all_fields:
            action_obj = action_cls.build_object(document_type,
                                                 field,
                                                 left_schema,
                                                 right_schema)
            if action_obj is not None:
                try:
                    left_schema = patch(action_obj.to_schema_patch(left_schema), left_schema)
                except (TypeError, ValueError, KeyError) as e:
                    raise ActionError(
                        f"Unable to apply schema patch of {action_obj!r}. More likely that the "
                        f"schema is corrupted. You can use schema repair tools to fix this issue"
                    ) from e

                yield action_obj


def get_all_document_types(left_schema: Schema, right_schema: Schema) -> Iterable[str]:
    """
    Return list of all document types collected from both schemas

    Document types should be processed in certain order depending
    if they are embedded or not, inherited or not. Embedded documents
    should be processed before common documents because their fields
    could refer to embedded documents which must be present in schema
    at the moment when migration of document is running. Also base
    document (common or embedded) should be processed before derived one

    So, common order is:

    1. Embedded documents
    2. Derived embedded documents ordered by their hierarchy depth
    3. Documents
    4. Derived documents ordered by their hierarchy depth

    :param left_schema:
    :param right_schema:
    :return:
    """
    def mark(document_type: str) -> int:
        # 0 -- embedded documents
        # 0 + derived depth -- embedded documents
        # 1000 -- documents
        # 1000 + derived_depth -- documents
        m = 0
        if not document_type.startswith(flags.EMBEDDED_DOCUMENT_NAME_PREFIX):
            m = 1000

        m += document_type.count(flags.DOCUMENT_NAME_SEPARATOR)

        return m

    all_document_types = ((k, mark(k)) for k in left_schema.keys() | right_schema.keys())
    all_document_types = sorted(all_document_types, key=lambda x: x[1])

    return [k[0] for k in all_document_types]



__all__ = [
    'actions_registry',
    'BaseActionMeta',
    'BaseAction',
    'BaseFieldAction',
    'BaseDocumentAction',
    'BaseCreateDocument',
    'BaseDropDocument',
    'BaseRenameDocument',
    'BaseAlterDocument'
]

import logging
import weakref
from abc import ABCMeta, abstractmethod
from copy import deepcopy
from typing import Dict, Type, Optional, Mapping, Any

from pymongo.database import Database

import mongoengine_migrate.flags as flags
from mongoengine_migrate.exceptions import ActionError, SchemaError
from mongoengine_migrate.fields.registry import type_key_registry
from mongoengine_migrate.schema import Schema
from mongoengine_migrate.utils import Diff, UNSET, document_type_to_class_name
from mongoengine_migrate.graph import MigrationPolicy
from mongoengine_migrate.updater import DocumentUpdater

#: Migration Actions registry. Mapping of class name and its class
actions_registry: Dict[str, Type['BaseAction']] = {}


log = logging.getLogger('mongoengine-migrate')


class BaseActionMeta(ABCMeta):
    def __new__(mcs, name, bases, attrs):
        attrs['_meta'] = weakref.proxy(mcs)

        c = super(BaseActionMeta, mcs).__new__(mcs, name, bases, attrs)
        if not name.startswith('Base'):
            actions_registry[name] = c

        return c


class BaseAction(metaclass=BaseActionMeta):
    """Base class for migration actions

    Action represents one change within migration such as field
    altering, collection renaming, collection dropping, etc.

    Migration file typically consists of actions following by each
    other. Every action accepts collection name and other parameters
    (if any) which describes change.

    Action also can be represented as dictdiff diff in order to apply
    schema changes.
    """

    #: Priority which this action will be tested with. The smaller
    #: priority number, the higher priority this action has.
    #: This flag is suitable for rename actions which should get tested
    #: before create/drop actions. Default is 5 which means normal
    priority = 12

    def __init__(self, document_type: str, *, dummy_action: bool = False, **kwargs):
        """
        :param document_type: Document type in schema which will
         Action will use to make changes
        :param dummy_action: If True then the action will not
         perform any queries on db during migration, but still used
         for changing the db schema
        :param kwargs: Action keyword parameters
        """
        self.document_type = document_type
        self.dummy_action = dummy_action
        self.parameters = kwargs
        self._run_ctx = None  # Run context, filled by `prepare()`

    def prepare(self, db: Database, left_schema: Schema, migration_policy: MigrationPolicy):
        """
        Prepare action before Action run (both forward and backward)
        :param db: pymongo.Database object
        :param left_schema: db schema before migration (left side)
        :param migration_policy:
        :return:
        """
        self._prepare(db, left_schema, migration_policy, True)

    def _prepare(self,
                 db: Database,
                 left_schema: Schema,
                 migration_policy: MigrationPolicy,
                 ensure_existence: bool):
        if ensure_existence and self.document_type not in left_schema:
            raise SchemaError(f'Document {self.document_type} does not exist in schema')
        elif not ensure_existence and self.document_type in left_schema:
            raise SchemaError(f'Document {self.document_type} already exists in schema')

        collection_name = self.parameters.get('collection')
        if not collection_name:
            docschema = left_schema.get(self.document_type)
            if docschema:
                collection_name = docschema.parameters.get('collection')

        collection = db[collection_name] if collection_name else db['COLLECTION_PLACEHOLDER']

        self._run_ctx = {
            'left_schema': left_schema,
            'db': db,
            'collection': collection,
            'migration_policy': migration_policy
        }

    def cleanup(self):
        """Cleanup after Action run (both forward and backward)"""

    @abstractmethod
    def run_forward(self):
        """
        DB commands to be run in forward direction.

        All queries executed here must be idempotental, i.e. give the
        same result after repeated execution. This is because if any
        query would fail then migration process will be aborted, and
        repeated migration run will execute the same commands in this
        case until the migration will get finished.
        """

    @abstractmethod
    def run_backward(self):
        """
        DB commands to be run in backward direction

        All queries executed here must be idempotental, i.e. give the
        same result after repeated execution. This is because if any
        query would fail then migration process will be aborted, and
        repeated migration run will execute the same commands in this
        case until the migration will get finished.
        """

    @abstractmethod
    def to_schema_patch(self, left_schema: Schema):
        """
        Return dictdiff patch should get applied in a forward direction
        run
        :param left_schema: schema state before the Action would get
         applied (left side)
        :return: schema diff
        """

    @abstractmethod
    def to_python_expr(self) -> str:
        """
        Return string of python code which creates current object with
        the same state
        """

    def __repr__(self):
        params_str = ', '.join(f'{k!r}={v!r}' for k, v in self.parameters.items())
        args_str = repr(self.document_type)
        if self.dummy_action:
            params_str += f', dummy_action={self.dummy_action}'
        return f'{self.__class__.__name__}({args_str}, {params_str})'

    def __str__(self):
        args_str = repr(self.document_type)
        if self.dummy_action:
            args_str += f', dummy_action={self.dummy_action}'
        return f'{self.__class__.__name__}({args_str}, ...)'


class BaseFieldAction(BaseAction):
    """
    Base class for action which affects on one field in a collection
    """

    def __init__(self, document_type: str, field_name: str, **kwargs):
        """
        :param document_type: collection name to be affected
        :param field_name: changing mongoengine document field name
        """
        super().__init__(document_type, **kwargs)
        self.field_name = field_name

        db_field = kwargs.get('db_field')
        if db_field and '.' in db_field:
            raise ActionError(f"db_field must not contain dots "
                              f"{self.document_type}.{self.field_name}")

    def get_field_handler_cls(self, type_key: str):
        """Concrete FieldHandler class for a given type key"""
        if type_key not in type_key_registry:
            raise SchemaError(f'Unknown type_key in {self.document_type}.{self.field_name}: '
                              f'{type_key}')

        return type_key_registry[type_key].field_handler_cls

    @classmethod
    @abstractmethod
    def build_object(cls,
                     document_type: str,
                     field_name: str,
                     left_schema: Schema,
                     right_schema: Schema) -> Optional['BaseFieldAction']:
        """
        Factory method which tests if current action type could process
        schema changes for a given collection and field. If yes then
        it produces object of current action type with filled out
        perameters. If no then it returns None.

        This method is used to guess which action is suitable to
        reflect schema change. It's called for several times for each
        field which was modified by a user in mongoengine documents.

        For example, on field deleting the method defined in
        CreateField action should return None, but those one in
        DropField action should return DeleteField object with
        filled out parameters of change (type of field, required flag,
        etc.)

        :param document_type: document type in schema to consider
        :param field_name: field name to consider
        :param left_schema: database schema before a migration
         would get applied (left side)
        :param right_schema: database schema after a migration
         would get applied (right side)
        :return: object of self type or None
        """
        pass

    def prepare(self, db: Database, left_schema: Schema, migration_policy: MigrationPolicy):
        self._prepare(db, left_schema, migration_policy, True)

        self._run_ctx['left_field_schema'] = \
            left_schema[self.document_type].get(self.field_name, {})

    def _prepare(self,
                 db: Database,
                 left_schema: Schema,
                 migration_policy: MigrationPolicy,
                 ensure_existence: bool):
        super()._prepare(db, left_schema, migration_policy, True)

        if ensure_existence and self.field_name not in left_schema[self.document_type]:
            raise SchemaError(f'Field {self.document_type}.{self.field_name} '
                              f'does not exist in schema')
        elif not ensure_existence and self.field_name in left_schema[self.document_type]:
            raise SchemaError(f'Field {self.document_type}.{self.field_name} '
                              f'already exists in schema')

    def to_python_expr(self) -> str:
        parameters = {
            name: getattr(val, 'to_python_expr', lambda: repr(val))()
            for name, val in self.parameters.items()
        }
        if self.dummy_action:
            parameters['dummy_action'] = True

        kwargs_str = ''.join(f", {name}={val}" for name, val in sorted(parameters.items()))
        return f'{self.__class__.__name__}({self.document_type!r}, {self.field_name!r}' \
               f'{kwargs_str})'

    def _get_field_handler(self, type_key: str, left_field_schema: dict, right_field_schema: dict):
        """
        Return FieldHandler object by type_key
        :param type_key: field type_key string
        :param left_field_schema: left schema which will be passed to
         a field
        :param right_field_schema: right schema which will be passed to
         a field
        :return: concrete FieldHandler object
        """
        handler_cls = self.get_field_handler_cls(type_key)
        handler = handler_cls(self._run_ctx['db'],
                              self.document_type,
                              self._run_ctx['left_schema'],
                              left_field_schema,
                              right_field_schema,
                              self._run_ctx['migration_policy'])
        return handler

    def __repr__(self):
        params_str = ', '.join(f'{k!r}={v!r}' for k, v in self.parameters.items())
        args_str = f'{self.document_type!r}, {self.field_name!r}'
        if self.dummy_action:
            params_str += f', dummy_action={self.dummy_action}'
        return f'{self.__class__.__name__}({args_str}, {params_str})'

    def __str__(self):
        args_str = f'{self.document_type!r}, {self.field_name!r}'
        if self.dummy_action:
            args_str += f', dummy_action={self.dummy_action}'
        return f'{self.__class__.__name__}({args_str}, ...)'


class BaseDocumentAction(BaseAction):
    """
    Base class for actions which change a document (collection or
    embedded document) at whole such as renaming, creating, dropping, etc.
    """
    @classmethod
    @abstractmethod
    def build_object(cls,
                     document_type: str,
                     left_schema: Schema,
                     right_schema: Schema) -> Optional['BaseDocumentAction']:
        """
        Factory method which tests if current action type could process
        schema changes for a given collection at whole. If yes then
        it produces object of current action type with filled out
        perameters. If no then it returns None.

        This method is used to guess which action is suitable to
        reflect schema change. It's called for several times for each
        collection which was modified by a user in mongoengine
        documents.

        For example, on collection deleting the method defined in
        CreateCollection action should return None, but those one in
        DropCollection action should return DropCollection object with
        filled out parameters of change (collection name, indexes, etc.)

        :param document_type: document type in schema to consider
        :param left_schema: database schema before a migration
         would get applied (left side)
        :param right_schema: database schema after a migration
         would get applied (right side)
        :return: object of self type or None
        """
        pass

    def to_python_expr(self) -> str:
        parameters = {
            name: getattr(val, 'to_python_expr', lambda: repr(val))()
            for name, val in self.parameters.items()
        }
        if self.dummy_action:
            parameters['dummy_action'] = True

        kwargs_str = ''.join(f", {name}={val}" for name, val in sorted(parameters.items()))
        return f'{self.__class__.__name__}({self.document_type!r}{kwargs_str})'

    def _is_my_collection_used_by_other_documents(self) -> bool:
        """Return True if some of documents uses the same collection"""
        docschema = self._run_ctx['left_schema'].get(self.document_type)
        if docschema:
            collection_name = docschema.parameters.get('collection')
        else:
            collection_name = self.parameters.get('collection')

        return collection_name and any(
            v.parameters.get('collection') == collection_name
            for k, v in self._run_ctx['left_schema'].items()
            if k != self.document_type and not k.startswith(flags.EMBEDDED_DOCUMENT_NAME_PREFIX)
        )


class BaseCreateDocument(BaseDocumentAction):
    @classmethod
    def build_object(cls, document_type: str, left_schema: Schema, right_schema: Schema):
        if document_type not in left_schema and document_type in right_schema:
            return cls(document_type=document_type, **right_schema[document_type].parameters)

    def to_schema_patch(self, left_schema: Schema):
        item = Schema.Document()
        item.parameters.update(self.parameters)
        return [('add', '', [(self.document_type, item)])]

    def prepare(self, db: Database, left_schema: Schema, migration_policy: MigrationPolicy):
        self._prepare(db, left_schema, migration_policy, False)


class BaseDropDocument(BaseDocumentAction):
    @classmethod
    def build_object(cls, document_type: str, left_schema: Schema, right_schema: Schema):
        if document_type in left_schema and document_type not in right_schema:
            return cls(document_type=document_type)  # FIXME: parameters (indexes, acl, etc.)

    def to_schema_patch(self, left_schema: Schema):
        item = left_schema[self.document_type]
        return [('remove', '', [(self.document_type, item)])]


class BaseRenameDocument(BaseDocumentAction):
    #: How much percent of items in schema diff of two collections
    #: should be equal to consider such change as collection rename
    #: instead of drop/create
    similarity_threshold = 70

    def __init__(self, document_type: str, *, new_name, **kwargs):
        super().__init__(document_type, new_name=new_name, **kwargs)
        self.new_name = new_name

    @classmethod
    def build_object(cls, document_type: str, left_schema: Schema, right_schema: Schema):
        # Check if field exists under different name in schema.
        # Field also can have small schema changes in the same time
        # So we try to get similarity percentage and if it more than
        # threshold then we're consider such change as rename/alter.
        # Otherwise it is drop/create
        match = document_type in left_schema and document_type not in right_schema
        if not match:
            return

        left_document_schema = left_schema[document_type]
        candidates = []
        matches = 0
        compares = 0
        for right_document_type, right_document_schema in right_schema.items():
            # Skip collections which was not renamed
            if right_document_type in left_schema:
                continue

            # Exact match, collection was just renamed
            if left_document_schema == right_document_schema:
                candidates = [(right_document_type, right_document_schema)]
                break

            # Try to find collection by its schema similarity
            # Compares are counted as every field schema comparing
            common_fields = left_document_schema.keys() | right_document_schema.keys()
            for field_name in common_fields:
                left_field_schema = left_document_schema.get(field_name, {})
                right_field_schema = right_document_schema.get(field_name, {})
                common_keys = left_field_schema.keys() & right_field_schema.keys()
                compares += len(common_keys)
                matches += sum(
                    left_field_schema[k] == right_field_schema[k]
                    for k in common_keys
                )

            if compares > 0 and (matches / compares * 100) >= cls.similarity_threshold:
                candidates.append((right_document_type, right_document_schema))

        if len(candidates) == 1:
            return cls(document_type=document_type, new_name=candidates[0][0])

    def to_schema_patch(self, left_schema: Schema):
        item = left_schema[self.document_type]
        return [
            ('remove', '', [(self.document_type, item)]),
            ('add', '', [(self.new_name, item)])
        ]


class BaseAlterDocument(BaseDocumentAction):
    @classmethod
    def build_object(cls, document_type: str, left_schema: Schema, right_schema: Schema):
        match = document_type in left_schema \
                and document_type in right_schema \
                and left_schema[document_type].parameters != right_schema[document_type].parameters
        if match:
            return cls(document_type=document_type, **right_schema[document_type].parameters)

    def to_schema_patch(self, left_schema: Schema):
        # Constructor args are written to parameters directly, not as
        # diff. Unlike the AlterField there are no schema skel here,
        # so we can't delete a parameter implicitly
        left_item = left_schema[self.document_type]
        right_item = deepcopy(left_item)
        right_item.parameters.clear()
        right_item.parameters.update(self.parameters)

        return [('change', self.document_type, (left_item, right_item))]

    def run_forward(self):
        self._run_migration(self._run_ctx['left_schema'][self.document_type],
                            self.parameters,
                            swap=False)

    def run_backward(self):
        self._run_migration(self._run_ctx['left_schema'][self.document_type],
                            self.parameters,
                            swap=True)

    def _run_migration(self,
                       self_schema: Schema.Document,
                       parameters: Mapping[str, Any],
                       swap: bool = False):
        # Try to process all parameters on same order to avoid
        # potential problems on repeated launches if some query on
        # previous launch was failed
        for name in sorted(parameters.keys() | self_schema.parameters.keys()):
            left_value = self_schema.parameters.get(name, UNSET)
            right_value = parameters.get(name, UNSET)
            if left_value == right_value:
                continue

            diff = Diff(
                old=right_value if swap else left_value,
                new=left_value if swap else right_value,
                key=name
            )

            log.debug(">> Change %s: %s => %s", repr(name), repr(diff.old), repr(diff.new))
            try:
                method = getattr(self, f'change_{name}')
            except AttributeError as e:
                raise SchemaError(f'Unknown document parameter: {name}') from e

            inherit = self._run_ctx['left_schema'][self.document_type].parameters.get('inherit')
            document_cls = document_type_to_class_name(self.document_type) if inherit else None
            updater = DocumentUpdater(self._run_ctx['db'], self.document_type,
                                      self._run_ctx['left_schema'], '',
                                      self._run_ctx['migration_policy'], document_cls)

            method(updater, diff)

    def _check_diff(self, diff: Diff, can_be_none=True, check_type=None):
        if diff.new == diff.old:
            raise SchemaError(f'Parameter {diff.key} does not changed from previous Action')

        if check_type is not None:
            if diff.old not in (UNSET, None) and not isinstance(diff.old, check_type) \
                    or diff.new not in (UNSET, None) and not isinstance(diff.new, check_type):
                raise SchemaError(f'{diff.key} must have type {check_type!r}')

        if not can_be_none:
            if diff.old is None or diff.new is None:
                raise SchemaError(f'{diff.key} could not be None')
